mod bootstrapping;
mod dghv;
mod dghv_asym;
mod utils;
use crate::dghv::{decrypt_bit, encrypt_bit, generate_secret_key};
use std::path::{Path, PathBuf};
use rug::Integer;
use clap::{Parser, Subcommand};


#[derive(Parser)]
#[command(author, version, about, long_about = None)]
#[command(propagate_version = true)]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    /// Generate a secret key
    Keygen {
        /// security parameter eta (p bit-length)
        #[clap(short = 's', long, default_value_t = 6400)]
        size: u32,
    },
    /// Encrypt a single bit
    Encrypt {
        /// bit to encrypt (0 or 1)
        #[clap(short = 'b', long)]
        bit: u8,
        /// secret key (hex)
        #[clap(short = 'k', long)]
        key: String,
        /// noise parameter
        #[clap(short = 'n', long, default_value_t = 80)]
        noise: u32,
    },
    /// Decrypt a single bit
    Decrypt {
        /// ciphertext to decrypt (hex)
        #[clap(short = 'c', long)]
        ciphertext: String,
        /// secret key (hex)
        #[clap(short = 'k', long)]
        key: String,
    },
}

fn main() {
    let cli = Cli::parse();

    match cli.command {
        Commands::Keygen { size } => {
            let sk = generate_secret_key(size as u32);
            println!("Secret key (p): {}", sk.to_string_radix(16));
        }
        Commands::Encrypt { bit, key, noise } => {
            let p = Integer::from_str_radix(&key, 16)
                .expect("Invalid secret key: must be hexadecimal integer");
            let ct = encrypt_bit(bit, &p, noise as u32, noise as u32);
            println!("Encrypted bit (c): {}", ct.to_string_radix(16));
        }
        Commands::Decrypt { ciphertext, key } => {
            let p = Integer::from_str_radix(&key, 16)
                .expect("Invalid secret key: must be hexadecimal integer");
            let ct = Integer::from_str_radix(&ciphertext, 16)
                .expect("Invalid ciphertext: must be hexadecimal integer");
            let res = decrypt_bit(&ct, &p);
            println!("Decrypted bit: {}", res);
        }
    }
}
